open Constants;;
(**
 A bounder for multibin instances.
 See help for details.

 @version March 2004
 @author Pierre Lemaire
*)

(* ========================================================================= *)
(* ========================================================================= *)

open Options;;
open Misc;;
open Item;;
open Instance;;
open Solution;;
open Solution_bf;;
open Solution_dbf;;
open Solution_abf;;
open Solution_dc;;
open Solution_adc;;

open Printf;;

let bounder() = try
  let time_init = Sys.time()
  and usage_msg = "usage is: "^Sys.executable_name^" [options] [inputfile]\navailable options are:"
  and bound     = ref "best"
  and sort      = ref ""
  and fvalname  = ref ""
  and dispval   = ref true
  and all       = ref false
  and dispheaders = ref false
  and fallname  = ref ""
  and inputname = ref ""
  and disptime  = ref false
  and ftimename = ref ""
  and seed      = ref (-1)
  and x         = ref (-1)
  in let arglist =
    [ ("-o",       Arg.String (set_id fvalname), "         where the values of the bounds are printed (default is stdout)");
      ("-display", Arg.String (set_display),     "(val) sets the mode [simple|normal|pretty] for displaying items (default is simple)");
      ("-val",     Arg.Set    (dispval) ,        "          turns bounds' value printing on (default)");
      ("-noval",   Arg.Clear  (dispval),         "        turns bounds' value printing off");

      ("-all",     Arg.Set    (all),             "          prints a summary of everything that's relevant");
      ("-fall",    Arg.String (set_id fallname), "         where '-all' is printed");
      ("-headers", Arg.Set    (dispheaders),     "      turns on the printing of headers for '-all'");
      ("-noheaders", Arg.Clear (dispheaders),    "    turns off the printing of the headers for '-all' (default)");

      ("-bound",   Arg.String (set_id bound),    "(val)    sets the bounds to compute [best|vol|vol+|hmax|nmin0|nmin|bf|dbf|abf|dc|adc] (default is best)");
      ("-sort",    Arg.String (set_id sort),     "(val)    sets the order [x1y1[x2y2]] where xi is the direction, in (i-ncreasing, d-ecreasing, a-lea) \n                 and yi is the criterion (when direction is not 'alea'), in (s-ize, h-eight, v-olume)");

      ("-time",    Arg.Set    (disptime),        "         turns running-time printing on");
      ("-notime",  Arg.Clear  (disptime),        "       turns running-time printing off (default)");
      ("-ftime",   Arg.String (set_id ftimename),"(name)  sets where the times are printed (default is stdout)");

      ("-trace",   Arg.String (set_trace),      "(val)   sets the degree of verbosity [none|some|most|all] of the output (default is none)");
      ("-ftrace",  Arg.String (set_ftrace),     "(name) sets where the trace is printed (default is stderr)");

      ("-seed",    Arg.Int    (set_id seed),    "(int)    sets the seed for random numbers");
      ("-x",       Arg.Int    (set_id x),       "(int)       sets the generic parameter x\n                 x may be: the number of generated solutions for abf");
    ]
  in
    Arg.parse arglist (set_id inputname) usage_msg;
    if (!seed = -1) then Random.self_init() else Random.init !seed;
    let fval  = open_file !fvalname  stdout
    and fall  = open_file !fallname  stdout
    and ftime = open_file !ftimename stdout
    and inst = tracer 1 "Read instance...\n"; read_instance !inputname
    and start_time = Sys.time()
  in
    (* ----- get items ----- *)
    if (String.length !sort) > 0 then begin
      try tracer 1 "Sort items...\n";
          set_compare_item !sort; inst#sort_items()
      with Invalid_argument msg -> fprintf stderr "warning: wrong order format: (ignored)\n"
    end ;
    let items = Array.of_list (inst#get_items())
    and m     = inst#get_m()
    in
      let n = Array.length items
  in
    tracer 1 "Compute bound...\n";
    (* ----- bounds ----- *)
    (* format: lb, ub [max/min_int ~ unset] *)

    (* LB: hmax *)
    let  b_hmax () = 
      tracer 2 "...hmax bound\n";
      (inst#get_hmax(), max_int)

    (* LB+UB: vol *)
    and b_vol () = 
      tracer 2 "...vol bound\n";
      let v = inst#get_volume()
      and h = inst#get_hmax()
      in (int_of_float (ceil ((float_of_int v) /. (float_of_int m))),
          h+int_of_float (ceil ((float_of_int v) /. (float_of_int m))))

    (* LB: nmin ; !!! the instance is modified !!! *)
    and b_nmin0 () =
      tracer 2 "...nmin0 bound\n";
      set_compare_item1 (cORDER_INCR + cORDER_H);
      Array.sort (compare_item) items;
      let s = ref 0
      in for i=0 to (n-1) do s := !s + items.(i)#get_size; done;
         let nmin = int_of_float (ceil ((float_of_int !s) /. (float_of_int m)))
         and h = ref 0
         in for i=0 to (nmin-1) do h := !h + items.(i)#get_height; done;
            (!h, max_int)

    (* LB: nmin modified ; !!! the instance is modified !!! *)
    and b_nmin () =
      tracer 2 "...nmin bound\n";
      set_compare_item1 (cORDER_INCR + cORDER_H);
      set_compare_item2 (cORDER_INCR + cORDER_S);
      Array.sort (compare_item) items;
      let s = ref 0
      in for i=0 to (n-1) do s := !s + items.(i)#get_size; done;
         let nmin = int_of_float (ceil ((float_of_int !s) /. (float_of_int m)))
         and t = m * (int_of_float (floor (float_of_int !s)/.(float_of_int m)))
         in let a = let i = ref 0
                    in while ((!s - items.(!i)#get_size > t) && (n - !i > nmin)) do
                         s := !s - items.(!i)#get_size;
                         i := !i + 1;
                       done;
                       !i
         and h = ref 0 
         in for i=a to a+nmin-1 do h := !h + items.(i)#get_height; done;
            (!h, max_int)

    (* L+(Lv) !!! the instance is modified !!! *)
    and b_Lv () =
      tracer 2 "...vol+ bound\n";
      set_compare_item1 (cORDER_DECR + cORDER_H);
      Array.sort (compare_item) items;
      let v = ref (inst#get_volume())
      and h = inst#get_hmax()
      in let l1 = max h (int_of_float (ceil ((float_of_int !v) /. (float_of_int m))))
         and u1 = h+(int_of_float (ceil ((float_of_int !v) /. (float_of_int m))))
      in let l = ref (max l1 (!v/m))
         and g = ref 0
         and p = ref (n-1)
      in while (!g <= !p) do
           let e=u1 - items.(!g)#get_height
           in if items.(!g)#get_height > l1 then begin
                v := !v + items.(!g)#get_size*(l1 - items.(!g)#get_height)
              end;
              g := !g +1;
              while !p > !g && (items.(!p)#get_height > e) do
                v := !v - items.(!g)#get_volume;
                p := !p - 1;
              done;
              l := max !l (int_of_float (ceil ((float_of_int !v) /. (float_of_int m))))
         done;
         (!l, max_int)

    (* LB+UB: bf *)
    and b_bf () = 
      tracer 2 "...bf bound\n";
      let v = let sol = new solution_bf m in sol#build inst; sol#get_value()
      and mf = float_of_int (inst#get_m())
      and hf = float_of_int (inst#get_hmax())
      in (v - (int_of_float (ceil ( (mf -. 1.0)*.hf /. mf))), v)

    (* LB+UB: dbf ; !!! the instance is modified !!! *)
    and b_dbf () = 
      tracer 2 "...dbf bound\n";
      set_compare_item1 (cORDER_DECR + cORDER_H); inst#sort_items();
      let v = let sol = new solution_bf m in sol#build inst; sol#get_value()
      in (int_of_float (ceil ((float_of_int (3*v)) /. 4.0)), v)

    (* LB+UB: dc *)
    and b_dc () = 
      tracer 2 "...dc bound\n";
      let v = let sol = new solution_dc m in sol#build inst; sol#get_value()
      and mf = float_of_int (inst#get_m())
      and hf = float_of_int (inst#get_hmax())
      in (v - (int_of_float (ceil ( (mf -. 1.0)*.hf /. mf))), v)

    (* LB+UB : abf ; !!! the instance is modified !!! *)
    and b_abf () =
      tracer 2 "...abf bound\n";
      set_compare_item "a";
      let x = if (!x > 0) then !x else 10
      and mf = float_of_int (inst#get_m())
      and hf = float_of_int (inst#get_hmax())
      and v = let sol = new solution_bf m in sol#build inst; sol#get_value()
      in let best  = ref v
      and    worst = ref v
      in for i=2 to x do
           inst#sort_items();
           let v = let sol = new solution_bf m in sol#build inst; sol#get_value()
           in if (v < !best)  then best  := v;
              if (v > !worst) then worst := v;
         done;
         (!worst - (int_of_float (ceil ( (mf -. 1.0)*.hf /. mf))), !best)

    (* LB+UB : adc ; !!! the instance is modified !!! *)
    and b_adc () =
      tracer 2 "...adc bound\n";
      set_compare_item "a";
      let x = if (!x > 0) then !x else 10
      and mf = float_of_int (inst#get_m())
      and hf = float_of_int (inst#get_hmax())
      and v = let sol = new solution_dc m in sol#build inst; sol#get_value()
      in let best  = ref v
      and    worst = ref v
      in for i=2 to x do
           inst#sort_items();
           let v = let sol = new solution_dc m in sol#build inst; sol#get_value()
           in if (v < !best)  then best  := v;
              if (v > !worst) then worst := v;
         done;
         (!worst - (int_of_float (ceil ( (mf -. 1.0)*.hf /. mf))), !best)
   


    (* LB+UB : best --- remark: abf and adc are not included in this bound.
              !!! the instance is modified !!! *)
    in let b_best () =
      tracer 2 "...best bound...\n";
      let (lb_hmax, ub_hmax) = b_hmax ()
      and (lb_vol,  ub_vol ) = b_vol  ()
      and (lb_bf,   ub_bf)   = b_bf   ()
      and (lb_dc,   ub_dc)   = b_dc   ()
      and (lb_nmin ,ub_nmin) = b_nmin () (* !!! re-order the instance ! *)
      and (lb_dbf,  ub_dbf)  = b_dbf  () (* !!! re-order the instance ! *)
      and (lb_Lv,   ub_Lv)   = b_Lv   () (* !!! re-order the instance ! *)
      in List.fold_left max min_int [lb_hmax; lb_vol; lb_bf; lb_dc; lb_nmin; lb_dbf; lb_Lv],
         List.fold_left min max_int [ub_hmax; ub_vol; ub_bf; ub_dc; ub_nmin; ub_dbf; ub_Lv]
  in
    (* ----- display ----- *)
    let (lb, ub) =  match !bound with
      | "hmax"  -> b_hmax ()
      | "vol"   -> b_vol  ()
      | "nmin0" -> b_nmin0()
      | "nmin"  -> b_nmin ()
      | "vol+"  -> b_Lv   ()
      | "bf"    -> b_bf   ()
      | "dbf"   -> b_dbf  ()
      | "dc"    -> b_dc   ()
      | "abf"   -> b_abf  ()
      | "adc"   -> b_adc  ()
      | "best"  -> b_best () 
      | _       -> failwith (sprintf "error: unknown bound %s (abort)\n" !bound); 
  in
    if (!dispval) then begin
      pretty_print_int fval "LB: Hmax* >= " lb;
      pretty_print_int fval "UB: Hmax* <= " (if ub=max_int then -1 else ub)
    end;
    if (!disptime) then fprintf ftime "Computing = %f\n" (Sys.time() -. start_time);
    if (!all) then begin
      if (!dispheaders) then
        fprintf fall "file ; n ; m ; hmax ; smax ; order ; bound ; LB ; UB ; time\n";
      fprintf fall "%s ; %d ; %d ; %d ; %d ; %s ; %s ; %d ; %d ; %f\n" !inputname (inst#get_n()) (inst#get_m()) (inst#get_hmax()) (inst#get_smax()) !sort !bound lb (if ub=max_int then -1 else ub) (Sys.time() -. start_time);
    end;

    (* ----- end ----- *)
    close_out options.ftrace;
    close_out ftime;
    close_out fall; 
    close_out fval;
    exit 0;
with 
  | Failure msg -> fprintf stderr "%s" msg; exit 1;
(*  | _           -> fprintf stderr "unexpected error (BUG)"; exit 99;*)
;;

(* ========================================================================= *)
if !Sys.interactive
then
  ()
else begin
  bounder()
end;;

(* ========================================================================= *)
(* ========================================================================= *)
