open Constants;;
(**
 A population of [gen_sols] for the genetic algorithm.

 @version December 2003
 @author Pierre Lemaire
*)

open Options;;
open Gen_sol;;
open Gen_sol_none;;

(** [gen_pop sols] builds a new population with [sols] as [gen_sols]. *)
class gen_pop (sols:gen_sol array) =
object (self)
  val mutable sols = sols
  val mutable best = (new gen_sol_none 0 [||] :> gen_sol)

  (* parameters *)
  val mutable cross_rate   = 0.5
  val mutable mutate_rate  = 0.6
  val mutable fitness      = cGEN_FIT_LINEAR
  val mutable max_fitness  = 100
  val mutable decr_fitness = 36
  val mutable reproduction = cGEN_REP_DELETE_ALL
  val mutable steady_tries = 10
  val mutable elitism      = false

  initializer 
    let b = ref 0
    in (* the best solution *)
       for i=1 to (Array.length sols)-1 do 
         let v = sols.(i)#get_value()
         in if v < sols.(!b)#get_value() then b := i;
       done;
       best <- sols.(!b);

  (** sets the cross over rate. *)
  method set_cross_rate x = cross_rate <- x;

  (** sets the mutation rate. *)
  method set_mutate_rate x = mutate_rate <- x;

  (** sets the max fitness. *)
  method set_fitness x = match(x) with
    | "window" -> fitness <- cGEN_FIT_WINDOW
    | "linear" -> fitness <- cGEN_FIT_LINEAR
    | _        -> raise (Invalid_argument "invalid fitness")

  (** sets the max fitness value. *)
  method set_max_fitness x = max_fitness <- x;

  (** sets the decr fitness value. *)
  method set_decr_fitness x = decr_fitness <- x;

  (** turns elitism on/off. *)
  method set_elitism x = elitism <- x;

  (** sets the steady_tries. *)
  method set_steady_tries x = steady_tries <- x;

  (** sets the max fitness. *)
  method set_reproduction x = match(x) with
    | "deleteall"   -> reproduction <- cGEN_REP_DELETE_ALL
    | "steadystate" -> reproduction <- cGEN_REP_STEADY_STATE
    | _             -> raise (Invalid_argument "invalid reproduction")


  (** @return the best found [solution]. *)
  method best_solution () = best#to_solution();

  (** @return the value of the best found [solution]. *)
  method best_value () = best#get_value();

  (* sets the fitnesses of the individuals, and return the total fitness *)
  method private linear_fitness () =
    Array.sort (compare_gen_sol) sols;
    let total_fitness = ref 0
    and v = ref max_fitness
    and x = ref (sols.(0)#get_value())
    in for i=0 to (Array.length sols)-1 do
         if (!v > 0) then begin
           if sols.(i)#get_value() > !x then begin
             x := sols.(i)#get_value();
             v := max 0 (!v - decr_fitness);
           end;
           total_fitness := !total_fitness + !v;
         end;
         sols.(i)#set_fitness(!v);
       done;
       !total_fitness

  (* sets the fitnesses of the individuals, and return the total fitness *)
  method private windowing () =
    let worst = let w = ref (sols.(0)#get_value())
                in for i=1 to (Array.length sols -1) do
                     w := max !w (sols.(i)#get_value())
                   done;
                   !w
    and total_fitness = ref 0 
    in for i=0 to Array.length sols - 1 do
       let v = worst + 1 - sols.(i)#get_value();
       in sols.(i)#set_fitness(v);
          total_fitness := !total_fitness + v;
       done;
       !total_fitness

  method private roulette_wheel tf = 
    let s1,s2 = 1+Random.int tf, 1+Random.int tf
    in let s1,s2 = min s1 s2, max s1 s2
    in let i = ref (-1) (* last explored sol *)
       and t = ref 0    (* partial sum from 0 to i *)
       and p1 = ref 0   (* first selected index *)
       and p2 = ref 0   (* second selected index *)
       in while (!t<s1) do i := !i + 1; t := !t + sols.(!i)#get_fitness(); done;
          p1 := !i;
          while (!t<s2) do i := !i + 1; t := !t + sols.(!i)#get_fitness(); done;
          p2 := !i;
          if (!p1 = !p2) then begin
            if (!p1 = Array.length sols -1) then p2 := 0
            else p2 := !p1 + 1;
          end;
          !p1, !p2

  (* builds a new generation and throws the previous one away *)
  method private delete_all () =
    let tf = match (fitness) with
               | cGEN_FIT_LINEAR -> self#linear_fitness()
               | cGEN_FIT_WINDOW -> self#windowing()
               | _              -> raise (Invalid_argument "invalid fitness")
    and childs = Array.make (Array.length sols) (new gen_sol_none 0 [||] :> gen_sol)
    and bestidx = ref 0
    in
      for i=0 to (Array.length sols)-1 do
        let (p1, p2) = self#roulette_wheel tf
        in childs.(i) <- sols.(p1)#clone();
          if (Random.float 1.0 < cross_rate)  then childs.(i)#cross_with (sols.(p2));
          if (Random.float 1.0 < mutate_rate) then childs.(i)#mutate();
          if childs.(i)#get_value() < childs.(!bestidx)#get_value() then bestidx := i;
      done;
      if elitism then begin
        let w  = ref (childs.(0)#get_value())
        and wi = ref 0
        in for i=1 to (Array.length sols)-1 do
             if !w < childs.(i)#get_value() then begin
               wi := i; w := childs.(i)#get_value()
             end;
           done;
           childs.(!wi) <- best;
      end;
      sols <- childs;
      if (sols.(!bestidx)#get_value() < best#get_value()) then best <- sols.(!bestidx)


  (* builds a new generation without (too much) duplicates *)
  method private steady_state () =
    for i=0 to (Array.length sols)-1 do
      let tf = match (fitness) with
                 | cGEN_FIT_LINEAR -> self#linear_fitness()
                 | cGEN_FIT_WINDOW -> self#windowing()
                 | _              -> raise (Invalid_argument "invalid fitness")
      and ok = ref steady_tries
      in while !ok > 0 do
        let (p1, p2) = self#roulette_wheel tf
        in let child = sols.(p1)#clone()
        in tracer 3 (Printf.sprintf "...... child %2d ; try %2d\n" i !ok);
           if (Random.float 1.0 < cross_rate)  then child#cross_with (sols.(p2));
           if (Random.float 1.0 < mutate_rate) then child#mutate();
           let w  = ref min_int
           and wi = ref (-1)
           and i = ref 0
           in while !i < Array.length sols do
                if (child#equals sols.(!i)) && (!ok <> 1) then begin
                  i := Array.length sols;
                  wi := -1
                end else begin
                  if !w < sols.(!i)#get_value() then begin
                    wi := !i; w := sols.(!i)#get_value()
                  end;
                end;
              i := !i + 1;
              done;
              if (!wi >= 0) then begin
                sols.(!wi) <- child; 
                if child#get_value() < best#get_value() then best <- child; 
                ok := 0;
              end;
         ok := !ok - 1;  
         done;
      done;
      if elitism then begin
        let w  = ref (sols.(0)#get_value())
        and wi = ref 0
        in for i=1 to (Array.length sols)-1 do
             if !w < sols.(i)#get_value() then begin
               wi := i; w := sols.(i)#get_value()
             end;
           done;
           sols.(!wi) <- best;
      end;

  (** turns this population into the next generation. *)
  method next_generation () =
    match (reproduction) with
      | cGEN_REP_DELETE_ALL   -> self#delete_all ()
      | cGEN_REP_STEADY_STATE -> self#steady_state ()
      | _        -> raise (Invalid_argument "invalid reproduction")

end
