open Constants;;
(**
 A solution for the genetic algorithm.

 @version December 2003
 @author Pierre Lemaire
*)

open Options;;
open Item;;
open Instance;;
open Solution;;

(** *)
class virtual gen_sol (m:int) (items:item array) =
object (self)

  val m = m;
  val mutable items = items
  val mutable value   = -1;
  val mutable fitness = -1;

  method virtual get_type: unit -> int

  (** @return the number of [bins] this [gen_sol] contains. *)
  method get_m () = m;

  (** @return the value of this [gen_sol] (that is the height of its highest [bin]. *)
  method get_value () =
    if (value >= 0) then value else begin self#compute_value(); value end;


  (** @return the fitness of this [gen_sol] *)
  method get_fitness () = fitness;

  (** sets the fitness of this [gen_sol] *)
  method set_fitness (v) = fitness <- v;

  (** (re-)computes the value of this [gen_sol]. *)
  method private virtual compute_value: unit -> unit

  (** @return a copy of this [gen_sol] *)
  method virtual clone: unit -> gen_sol

  (** @return the [instance] this [gen_sol] is a solution of. *)
  method to_instance () =
    let inst = new instance m
    in for i=(Array.length items)-1 downto 0 do inst#add items.(i); done;
       inst;

  (** @return the [solution] this [gen_sol] is. *)
  method virtual to_solution: unit -> solution

  (** @return the list of the items *)
  method item_list () = Array.to_list items;

  (** @return [true] iff this [gen_sol] is equal to the given one. *)
  method equals (x:gen_sol) =
    let rec eq x y = match (x,y) with
      | ([],[]) -> true
      | (_,[])  -> false
      | ([],_)  -> false
      | (x::xs,y::ys) ->    (x#get_size   = y#get_size)
                         && (x#get_height = y#get_height)
                         && (eq xs ys)
    in    (self#get_type() = x#get_type())
       && (m = x#get_m())
       && (eq (self#item_list()) (x#item_list()));

  (** mutates this [gen_sol] (permuts a sub-list). *)
  method mutate () =
    tracer 2 "......mutation\n";
    let i,j = Random.int (Array.length items), Random.int (Array.length items)
    in let i,j = min i j, max i j
    in let tmp = Array.make (j-i+1) (new item 0 0)
    and per = Myrand.permutation (j-i+1)
    in for k=0 to (j-i) do tmp.(k) <- items.(i+per.(k)) done;
       for k=0 to (j-i) do items.(i+k) <- tmp.(k) done;
       value <- (-1)

  (** crosses this [gen_sol] with the given [gen_sol]. *)
  method cross_with (x:gen_sol) =
    tracer 2 "......cross over\n";
    let rec del x l = match (l) with
      | []     -> []
      | (y::ys) -> if (y=x) then ys else (y::del x ys)
    in let tmp = Array.make (Array.length items) (new item 0 0)
    and    lx  = ref (x#item_list())
    in (* select from this sol *)
       for i=0 to (Array.length tmp)-1 do
         if (Random.bool()) then begin
           tmp.(i) <- items.(i);
           lx := del items.(i) !lx
         end
       done;
       (* select from x *)
       for i=0 to (Array.length tmp)-1 do
         if (tmp.(i)#get_volume = 0) then begin
           tmp.(i) <- List.hd !lx;
           lx := List.tl !lx
         end
       done;
       items <- tmp;
       value <- (-1)


  (** prints this [gen_sol]. *)
  method print () =
    let print_item (x:item) = x#print
    in String.concat " " (List.map print_item (Array.to_list items))
end

let compare_gen_sol (x:gen_sol) (y:gen_sol) = x#get_value() - y#get_value();
