open Constants;;
(**
 A packer for multibin instances, based on a genetic algorithm.
 See help for details.

 @version December 2003
 @author Pierre Lemaire
*)

(* ========================================================================= *)
(* ========================================================================= *)

open Options;;
open Misc;;
open Instance;;
open Solution;;
open Genetic;;

open Printf;;

let genpacker() = try
  let time_init = Sys.time()
  and usage_msg = "usage is: "^Sys.executable_name^" [options] [inputfile]\navailable options are:"
  and disptime  = ref false
  and ftimename = ref ""
  and dispsol   = ref true
  and dispval   = ref true
  and dispbound = ref false
  and fsolname  = ref ""
  and algo      = ref "bf"
  and all       = ref false
  and dispheaders = ref false
  and fallname  = ref ""
  and inputname = ref ""
  and seed      = ref (-1)
  and pop       = ref 35
  and gen       = ref 100
  and crossrate = ref (-1.0)
  and mutaterate= ref (-1.0)
  and fitness   = ref ""
  and fitmax    = ref (-1)
  and fitdecr   = ref (-1)
  and repro     = ref ""
  and elitism   = ref false
  and steadtry  = ref 10
  in let arglist =
    [ ("-o",       Arg.String (set_id fsolname), "(name)       sets where the solution is printed (default is stdout)");
      ("-sol",     Arg.Set    (dispsol) ,        "           turns solution printing on (default)");
      ("-nosol",   Arg.Clear  (dispsol),         "         turns solution printing off");
      ("-display", Arg.String (set_display),     "(val)  sets the mode [simple|normal|pretty] for displaying items (default is simple)");
      ("-val",     Arg.Set    (dispval) ,        "           turns solution-value printing on (default)");
      ("-noval",   Arg.Clear  (dispval),         "         turns solution-value printing off");
      ("-lb",      Arg.Set    (dispbound),       "            turns lower bound printing on");
      ("-nolb",    Arg.Clear  (dispbound),       "          turns lower bound printing off (default)");

      ("-all",     Arg.Set    (all),             "           prints a summary of everything that's relevant");
      ("-fall",    Arg.String (set_id fallname), "          where '-all' is printed");
      ("-headers", Arg.Set    (dispheaders),     "       turns on the printing of headers for '-all'");
      ("-noheaders", Arg.Clear (dispheaders),    "     turns off the printing of the headers for '-all' (default)");

      ("-algo",    Arg.String (set_id algo),     "(val)     sets the algorithm [bf|dc] for computing the solution (default is bf)");

      ("-time",    Arg.Set    (disptime),        "          turns running-time printing on");
      ("-notime",  Arg.Clear  (disptime),        "        turns running-time printing off (default)");
      ("-ftime",   Arg.String (set_id ftimename),"(name)   sets where the times are printed (default is stdout)");

      ("-trace",   Arg.String (set_trace),      "(val)    sets the degree of verbosity [none|some|most|all] of the output (default is none)");
      ("-ftrace",  Arg.String (set_ftrace),     "(name)  sets where the trace is printed (default is stderr)");

      ("-seed",    Arg.Int    (set_id seed),    "(int)     sets the seed for random numbers");

      ("-pop",     Arg.Int    (set_id pop),     "(int)      sets the population");
      ("-gen",     Arg.Int    (set_id gen),     "(int)      the number of generations");
      ("-cross",   Arg.Float  (set_id crossrate), "(float)  the rate of cross overs");
      ("-mutate",  Arg.Float  (set_id mutaterate),"(float) the rate of mutations");
      ("-fitness", Arg.String (set_id fitness), "(val)  sets the fitness method [linear,window] (default is linear)");
      ("-fitmax",  Arg.Int    (set_id fitmax),  "(val)   sets the maximal fitness for linear method (default is 100)");
      ("-fitdecr", Arg.Int    (set_id fitdecr), "(val)  sets the decrease of the fitness for linear method (default is 20)");
      ("-repro",  Arg.String (set_id repro),    "(val)    sets the reproduction mode [deleteall|steadystate]");
      ("-elitism",   Arg.Set (elitism),         "       turns on the elitism strategy");
      ("-noelitism", Arg.Clear (elitism),         "     turns off the elitism strategy (default)");
      ("-steadytries", Arg.Int    (set_id steadtry), "(val)  sets the number of tries for non duplicate in steady-state reproduction (default is 10)");
    ]
  in
    Arg.parse arglist (set_id inputname) usage_msg;
    if (!seed = -1) then Random.self_init() else Random.init !seed;
    let fsol  = open_file !fsolname  stdout
    and ftime = open_file !ftimename stdout
    and fall  = open_file !fallname  stdout
    and time_reading = Sys.time()
    and inst = read_instance !inputname
    and time_computing = Sys.time()
    in
      let sol = genetic inst !algo !pop !gen
                        !crossrate !mutaterate
                        !fitness !fitmax !fitdecr
                        !repro !elitism !steadtry
      in
        if (!dispsol)   then begin fprintf fsol "%s\n" (sol#print()) end;
        if (!dispval)   then begin pretty_print_int fsol "VAL: Hmax   = " (sol#get_value()) end;
        if (!dispbound) then begin pretty_print_int fsol "LB:  Hmax* >= " (sol#get_bound()); end;
        if (!disptime) then begin
          fprintf ftime "Total     = %f\n" (Sys.time() -. time_init);
          fprintf ftime "Init      = %f\n" (time_reading -. time_init);
          fprintf ftime "Reading   = %f\n" (time_computing -. time_reading);
          fprintf ftime "Computing = %f\n" ((Sys.time ()) -. time_computing)
        end;
        if (!all) then begin
          if (!dispheaders) then begin
            fprintf fall "file ; n ; m ; hmax ; smax ; order ; algo ; value ; best LB ; reading ; sorting ; algorithm ; overall\n"
          end;
            fprintf fall "%s ; %d ; %d ; %d ; %d ; %s ; %s ; %d ; %d ; %f ; %f ; %f ; %f\n"
                         !inputname (inst#get_n()) (inst#get_m()) (inst#get_hmax()) (inst#get_smax())
                         "-" ("gen_"^(!algo)) (sol#get_value()) (sol#get_bound())
                         0.0 (time_computing -. time_reading) ((Sys.time ()) -. time_computing) (Sys.time() -. time_init)
        end;
        close_out options.ftrace;
        close_out fsol;
        close_out ftime;
        close_out fall;
        exit 0;
with 
  | Failure msg -> fprintf stderr "%s" msg; exit 1;
(*  | _           -> fprintf stderr "unexpected error (BUG)"; exit 99;*)
;;

(* ========================================================================= *)
if !Sys.interactive
then
  ()
else begin
  genpacker()
end;;

(* ========================================================================= *)
(* ========================================================================= *)
