(**
 The [instance] class, which represents an instance of a multibin packing.

 @version November 2003 - December 2003
 @author Pierre Lemaire
*)

open Item;;
open Bin;;

(** [instance m] creates a new empty [instance] of [m bins]. *)
class instance:
  int -> (* m *)
object
  val m: int
  val mutable value: int
  val mutable n: int
  val mutable hmax: int
  val mutable smax: int
  val mutable items: item list

  (** @return the number of [items] this [instance] contains. *)
  method get_n: unit -> int

  (** @return the number of [bins] this [instance] contains. *)
  method get_m: unit -> int

  (** @return the height of the highest [item] this [instance] contains. *)
  method get_hmax: unit -> int

  (** @return the height of the lowest [item] this [instance] contains. *)
  method get_hmin: unit -> int

  (** @return the size of the largest [item] this [instance] contains. *)
  method get_smax: unit -> int

  (** @return the size of the smallest [item] this [instance] contains. *)
  method get_smin: unit -> int

  (** @return the total volume of this [instance]. *)
  method get_volume: unit -> int

  (** @return the best of the volume-based or the hmax lower bound. *)
  method get_bound: unit -> int

  (** [add x] adds [item x] to this [instance]. *)
  method add: item -> unit

  (** returns the [items] of this [instance]. *)
  method get_items: unit -> item list

  (** sorts the [items] of this instance according to the current value of
      [compare_item]
      @see <Item.html> [Item.compare_item], [Item.set_compare_item].
   *)
  method sort_items: unit -> unit

  (** prints this [instance]. *)
  method print: unit -> string

end

(** [generate_instance m n hmin hmax vh smin smax vs algo seed] generates a new
    instance of [n items] and [m bins]. The [items] are generated according
    to the method [algo] (see below), and [seed] is used to initiate the
    pseudo-random number generator. The possible methods to generate an
    instance are:
{ul {- GEN_RAND:
         height is uniformly chosen in \[[hmin];[hmax]\];
      	 size is uniformly chosen in \[[smin];[smax]\]. }
    {- GEN_GAUSS:
         height is normaly chosen in \[[hmin];[hmax]\] with variance [vh];
      	 size is normaly chosen in \[[smin];[smax]\] with variance [vs]. }
    {- GEN_FIXED_V:
         height is [hmax];
         size is [smax].}
    {- GEN_FIXED_H:
         height is [hmax];
      	 size is uniformly chosen in \[[smin];[smax]\]. }
    {- GEN_FIXED_S:
         height is uniformly chosen in \[[hmin];[hmax]\];
      	 size is [smax].}
    {- GEN_STAIRS_H:
         height increases from [hmin] to [hmax] (then reduces back to [hmin]) by step of 1;
         size is uniformly chosen in \[[smin];[smax]\]. }
    {- GEN_STAIRS_S:
         height is uniformly chosen in \[[hmin];[hmax]\];
         size increases from [smin] to [smax] (then reduces back to [smin]) by step of 1. }
    {- GEN_STAIRS_V:
         height increases from [hmin] to [hmax] (then reduces back to [hmin]) by step of 1;
         size increases from [smin] to [smax] (then reduces back to [smin]) by step of 1. }
    {- GEN_DIFFERENCE:
         for an odd item: height is [h], uniformly chosen in \[[hmin];[hmax]\];
                          size is [s], uniformly chosen in \[1;[smax/2]\].
         for the following (even) item: height is [h];
                                        size is [m-s].}
    {- GEN_COMPLEMENT:
         for an odd item: height is  uniformly chosen in \[[hmin];[hmax]\];
                          size is [s], uniformly chosen in \[1;[smax/2]\].
         for the following (even) item: height is  uniformly chosen in \[[hmin];[hmax]\];
                                        size is [m-s].}
  }
    @raise Invalid_argument if the method [algo] is unknown.
*)
val generate_instance: int -> int -> int -> int -> int -> int -> int -> int -> int -> int -> instance

(** reads an [instance] from the specified file ("" for stdin) *)
val read_instance: string -> instance
