open Constants;;
(**
 A packer for multibin instances, with several fast heuristics.
 See help for details.

 @version November 2003 - April 2004
 @author Pierre Lemaire
*)

(* ========================================================================= *)
(* ========================================================================= *)

open Options;;
open Misc;;
open Item;;
open Instance;;
open Solution;;
open Solution_bf;;
open Solution_dbf;;
open Solution_abf;;
open Solution_pbf;;
open Solution_dc;;
open Solution_adc;;
open Solution_nb;;
open Solution_dnb;;
open Solution_op;;
open Solution_opv;;
open Solution_opx;;
open Solution_dop;;
open Solution_dopv;;
open Solution_dopx;;

open Printf;;

let packer() = try
  let time_init = Sys.time()
  and usage_msg = "usage is: "^Sys.executable_name^" [options] [inputfile]\navailable options are:"
  and disptime  = ref false
  and ftimename = ref ""
  and dispsol   = ref true
  and dispval   = ref true
  and dispbound = ref false
  and fsolname  = ref ""
  and algo      = ref "dbf"
  and sort      = ref ""
  and all       = ref false
  and dispheaders = ref false
  and fallname  = ref ""
  and inputname = ref ""
  and seed      = ref (-1)
  and x         = ref (-1)
  in let arglist =
    [ ("-o",       Arg.String (set_id fsolname), "(name)      sets where the solution is printed (default is stdout)");
      ("-sol",     Arg.Set    (dispsol) ,        "          turns solution printing on (default)");
      ("-nosol",   Arg.Clear  (dispsol),         "        turns solution printing off");
      ("-display", Arg.String (set_display),     "(val) sets the mode [simple|normal|pretty] for displaying items (default is simple)");
      ("-val",     Arg.Set    (dispval) ,        "          turns solution-value printing on (default)");
      ("-noval",   Arg.Clear  (dispval),         "        turns solution-value printing off");
      ("-lb",      Arg.Set    (dispbound),       "           turns lower bound printing on");
      ("-nolb",    Arg.Clear  (dispbound),       "         turns lower bound printing off (default)");

      ("-all",     Arg.Set    (all),             "          prints a summary of everything that's relevant");
      ("-fall",    Arg.String (set_id fallname), "         where '-all' is printed");
      ("-headers", Arg.Set    (dispheaders),     "      turns on the printing of headers for '-all'");
      ("-noheaders", Arg.Clear (dispheaders),    "    turns off the printing of the headers for '-all' (default)");

      ("-algo",    Arg.String (set_id algo),     "(val)    sets the algorithm [bf|dbf|abf|pbf|dc|adc|nb|dnb|op|opv|opx|dop|dopv|dopx] for computing the solution (default is dbf)");
      ("-sort",    Arg.String (set_id sort),     "(val)    sets the order [x1y1[x2y2]] where xi is the direction, in (i-ncreasing, d-ecreasing, a-lea) \n                 and yi is the criterion (when direction is not 'alea'), in (s-ize, h-eight, v-olume)");


      ("-time",    Arg.Set    (disptime),        "         turns running-time printing on");
      ("-notime",  Arg.Clear  (disptime),        "       turns running-time printing off (default)");
      ("-ftime",   Arg.String (set_id ftimename),"(name)  sets where the times are printed (default is stdout)");

      ("-trace",   Arg.String (set_trace),      "(val)   sets the degree of verbosity [none|some|most|all] of the output (default is none)");
      ("-ftrace",  Arg.String (set_ftrace),     "(name) sets where the trace is printed (default is stderr)");

      ("-seed",    Arg.Int    (set_id seed),    "(int)    sets the seed for random numbers");
      ("-x",       Arg.Int    (set_id x),       "(int)       sets the generic parameter x\n                 x may be: the number of generated solutions for abf");
    ]
  in
    Arg.parse arglist (set_id inputname) usage_msg;
    if (!seed = -1) then Random.self_init() else Random.init !seed;
    let fsol  = open_file !fsolname  stdout
    and ftime = open_file !ftimename stdout
    and fall  = open_file !fallname  stdout
    and time_reading = Sys.time()
    and inst = read_instance !inputname
    and time_sorting = Sys.time()
    in
      if (String.length !sort) > 0 then begin
        try set_compare_item !sort; inst#sort_items()
        with Invalid_argument msg -> fprintf stderr "warning: wrong order format: (ignored)\n"
      end
      else sort := "-";
      let sol = begin match (!algo) with
          | "bf"  -> (new solution_bf  (inst#get_m()) :> solution); 
          | "dbf" -> (new solution_dbf (inst#get_m()) :> solution);
          | "abf" -> (new solution_abf (inst#get_m()) (if (!x>0) then !x else 10) :> solution);
          | "pbf" -> (new solution_pbf (inst#get_m()) :> solution);
          | "dc"  -> (new solution_dc  (inst#get_m()) :> solution);
          | "adc" -> (new solution_adc (inst#get_m()) (if (!x>0) then !x else 10) :> solution);
          | "nb"  -> (new solution_nb  (inst#get_m()) :> solution); 
          | "dnb" -> (new solution_dnb (inst#get_m()) :> solution);
          | "op"  -> (new solution_op  (inst#get_m()) :> solution); 
          | "dop" -> (new solution_dop (inst#get_m()) :> solution);
          | "opv" -> (new solution_opv (inst#get_m()) :> solution); 
          | "dopv"-> (new solution_dopv(inst#get_m()) :> solution);
          | "opx" -> (new solution_opx (inst#get_m()) :> solution); 
          | "dopx"-> (new solution_dopx(inst#get_m()) :> solution);
          | _     -> failwith (sprintf "error: unknown algorithm %s (abort)\n" !algo); 
        end ; 
      and time_computing = Sys.time()
      in
        sol#build inst; 
        if (!dispsol) then begin   fprintf fsol "%s\n" (sol#print()) end;
        if (!dispval) then begin   pretty_print_int fsol "VAL: Hmax   = " (sol#get_value()) end;
        if (!dispbound) then begin pretty_print_int fsol "LB:  Hmax* >= " (sol#get_bound()); end;
        if (!disptime) then begin
          fprintf ftime "Total     = %f\n" (Sys.time() -. time_init);
          fprintf ftime "Init      = %f\n" (time_reading -. time_init);
          fprintf ftime "Reading   = %f\n" (time_sorting -. time_reading);
          fprintf ftime "Sorting   = %f\n" (time_computing -. time_sorting);
          fprintf ftime "Computing = %f\n" ((Sys.time ()) -. time_computing)
        end;
        if (!all) then begin
          if (!dispheaders) then begin
            fprintf fall "file ; n ; m ; hmax ; smax ; order ; algo ; value ; best LB ; reading ; sorting ; algorithm ; overall\n"
          end;
            fprintf fall "%s ; %d ; %d ; %d ; %d ; %s ; %s ; %d ; %d ; %f ; %f ; %f ; %f\n"
                         !inputname (inst#get_n()) (inst#get_m()) (inst#get_hmax()) (inst#get_smax())
                         !sort !algo (sol#get_value()) (sol#get_bound())
                         (time_sorting -. time_reading) (time_computing -. time_sorting) ((Sys.time ()) -. time_computing) (Sys.time() -. time_init)
        end;
        close_out options.ftrace;
        close_out fsol;
        close_out ftime;
        close_out fall;
        exit 0;
with 
  | Failure msg -> fprintf stderr "%s" msg; exit 1;
(*  | _           -> fprintf stderr "unexpected error (BUG)"; exit 99;*)
;;

(* ========================================================================= *)
if !Sys.interactive
then
  ()
else begin
  packer()
end;;

(* ========================================================================= *)
(* ========================================================================= *)
