open Constants;;
(**
 The [solution] class, which represents a packing of multibin objects.

 @version December 2003
 @author Pierre Lemaire
*)


open Options;;
open Item;;
open Bin;;
open Instance;;

(* hidden *)
let print_bin (x:bin) = x#print

exception Incompatible_solution;;

(** [solution m] creates a new empty [solution] of [m bins]. *)
class virtual solution (m:int) =
object (self)

  (** the value of this [solution]. *)
  val mutable value  = 0;

  (** the total volume of this [solution]. *)
  val mutable volume = 0;

  (** the height of the highest [item] in this [solution]. *)
  val mutable hmax   = 0;

  val mutable bins = Array.make m (new bin);

  initializer for i=0 to m-1 do bins.(i) <- new bin done;

  (** @return the number of [bins] this [solution] contains. *)
  method get_m () = Array.length bins;

  (** @return the value of this [solution] (that is the height of its highest [bin]. *)
  method get_value () = value;

  (** @return the maximal difference of height between two bins of this [solution]. *)
  method private get_delta () =
    let min = ref value 
    in for i=0 to (Array.length bins -1) do
         if (!min > bins.(i)#get_height) then min := bins.(i)#get_height
       done;
       value - !min

  (** @return the best known lower bound for this [solution]. \[Unless
      redefined, this bound is the best between the volume-bound, hmax, and
      the bound given by H(S) <= H(OPT) + (m-1)/m * Delta.\] *)
  method get_bound () =
    let fm  = float_of_int m
    and fvo = float_of_int volume
    and fva = float_of_int value
    and fd  = float_of_int (self#get_delta())
    in max hmax
           (max (int_of_float (ceil (fvo /. fm)))
                (int_of_float (ceil (fva -. fd*.(fm-.1.0)/.fm))))

  (** [add x i] adds [item x] to the [i]th [bin] of this [solution]. *)
  method private add (x:item) (i:int) =
    bins.(i)#add x;
    if (value < bins.(i)#get_height) then value <- bins.(i)#get_height;
    volume <- volume + x#get_height

  (** builds a [solution] with the given [instance].
      @raise Incompatible_solution if the [instance] is not compatible with this [solution]. *)
  method virtual build: instance -> unit

  (** prints this [solution]. *)
  method print () =
    String.concat "\n" (List.map print_bin (Array.to_list bins))
         
end
